/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): Romain Lenglet
 */

package org.objectweb.dream.cache;

import org.objectweb.dream.Pull;
import org.objectweb.dream.PullException;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.time.GetTimeStamp;
import org.objectweb.fractal.fraclet.annotation.annotations.Attribute;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;

/**
 * This class defines a basic cache. A cached message is considered as
 * up-to-date as long as it has not been cached more than <code>timeout</code>
 * milliseconds.
 * <p>
 * <code>timeout</code> can be changed using the
 * <code>CacheAttributeController</code> controller.
 */
@DreamComponent(controllerDesc = "dreamPrimitive")
@Provides(interfaces = { @Interface(name = "out-pull", signature = Pull.class) })
public class Cache implements Pull {

    /** The cache time out. */
    @Attribute(argument = "timeOut")
    protected long timeOut = 0;

    /** The message manager client interface. */
    @Requires(name = "message-manager")
    protected MessageManagerType messageManagerItf;

    /** The GetTimeStamp client interface. */
    @Requires(name = GetTimeStamp.ITF_NAME)
    protected GetTimeStamp getTimeStamp;

    /** The Pull client interface. */
    @Requires(name = "in-pull")
    protected Pull inPull;

    /** The last cached message. */
    protected Message cachedMessage;

    /** The last time stamp. */
    protected long lastTimeStamp;

    // -------------------------------------------------------------------------
    // Implementation of the Pull interface
    // -------------------------------------------------------------------------

    /**
     * @see Pull#pull()
     */
    public Message pull() throws PullException {
        long currentTimeStamp = getTimeStamp.getTimeStamp();
        if (currentTimeStamp - lastTimeStamp > timeOut) {
            lastTimeStamp = currentTimeStamp;
            if (cachedMessage != null) {
                messageManagerItf.deleteMessage(cachedMessage);
            }
            cachedMessage = inPull.pull();
        }
        return messageManagerItf.duplicateMessage(cachedMessage, true);
    }
}
