/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.channel;

import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.protocol.ExportIdentifierChunk;
import org.objectweb.dream.protocol.IPExportIdentifier;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.Attribute;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.Logger;

/**
 * This message transformer adds a {@link ExportIdentifierChunk} containing an
 * {@link IPExportIdentifier}. <br>
 * The name of the added chunk is specified using
 * <code> AddIPExportIdChunkAttributeController#setChunkName(String)</code>.<br>
 * The inet address must be specified using the
 * <code> AddIPExportIdChunkAttributeController#setHostname(String)</code>
 * method.
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
@Provides(interfaces = { @Interface(name = "in-push", signature = Push.class) })
public class AddIPExportIdChunk implements Push {
    // --------------------------------------------------------------------------
    // Fields
    // --------------------------------------------------------------------------

    protected IPExportIdentifier exportIdentifier;
    protected ChunkFactoryReference<ExportIdentifierChunk> exportIdChunkFactory;

    // ---------------------------------------------------------------------------
    // Attributes
    // ---------------------------------------------------------------------------

    /**
     * the name of added chunk
     */
    @Attribute(argument = "chunkName")
    protected String chunkName;
    /**
     * the hostname put in chunk
     */
    @Attribute(argument = "hostname")
    protected String hostname;
    /**
     * the port number put in chunk
     */
    @Attribute(argument = "port")
    protected int port;

    // ---------------------------------------------------------------------------
    // Client interfaces
    // ---------------------------------------------------------------------------

    @Requires(name = "out-push")
    protected Push outPushItf;
    @Requires(name = "message-manager")
    protected MessageManagerType messageManagerItf;

    // --------------------------------------------------------------------------
    // Services interfaces
    // --------------------------------------------------------------------------

    /**
     * Component reference
     */
    @Service
    Component weaveableC;

    /**
     * Logger of the component
     */
    @DreamMonolog()
    protected Logger logger;

    // ---------------------------------------------------------------------------
    // Implementation of Push interface
    // ---------------------------------------------------------------------------

    /**
     * @see Push#push(Message)
     */
    public void push(Message message) throws PushException {
        if (this.exportIdentifier == null) {
            this.exportIdentifier = new IPExportIdentifier(this.hostname, this.port);
        }
        if (this.exportIdChunkFactory == null) {
            this.exportIdChunkFactory = this.messageManagerItf
                    .getChunkFactory(ExportIdentifierChunk.class);
        }
        final ExportIdentifierChunk chunk = this.messageManagerItf
                .createChunk(this.exportIdChunkFactory);
        chunk.setExportIdentifier(this.exportIdentifier);
        this.messageManagerItf.addChunk(message, this.chunkName, chunk);
        this.outPushItf.push(message);
    }

    // ---------------------------------------------------------------------------
    // Overriding of AddIPExportIdChunkAttributeController interface
    // generated by fraclet
    // ---------------------------------------------------------------------------

    /**
     * Set the hostname
     */
    public void setHostname(String hostname) {
        this.hostname = hostname;
        this.exportIdentifier = null;
    }

    /**
     * Set the port
     */
    public void setPort(int port) {
        this.port = port;
        this.exportIdentifier = null;
    }
}
