/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.channel;

import org.objectweb.dream.IOPushException;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.protocol.messagePassing.MessagePassingOutgoingPush;
import org.objectweb.dream.protocol.messagePassing.MessagePassingProtocol;
import org.objectweb.dream.util.LoggerUtil;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;

/**
 * ChannelOut implementation.
 */
@DreamComponent(controllerDesc = "dreamPrimitive", needAsyncStart = true)
@Provides(interfaces = { @Interface(name = "in-push", signature = Push.class) })
public class ChannelOutImpl implements Push, ChannelInExportId {
    // ------------------------------------------------------------------------
    // ---
    // Fields
    // ------------------------------------------------------------------------
    // ---

    /**
     * Interface on which messages are transmitted to the lower protocol. This
     * field requires the "dreamPrimitive" controller descriptor because calls
     * on the "in-push" server interfaces must be blocked until the component is
     * started.
     */
    protected MessagePassingOutgoingPush lowerOutgoingPush;

    // ------------------------------------------------------------------------
    // ---
    // Client interfaces
    // ------------------------------------------------------------------------
    // ---

    /**
     * Interface used during startup to export lower protocol (i.e. get the
     * reference {@link #lowerOutgoingPush})
     */
    @Requires(name = "lower-protocol")
    protected MessagePassingProtocol lowerProtocolItf;

    /**
     * Resolver used to extract the {@link ExportIdentifier} from the pushed
     * message.
     */
    @Requires(name = DestinationResolver.ITF_NAME)
    protected DestinationResolver destinationResolverItf;

    // ------------------------------------------------------------------------
    // --
    // Services interfaces
    // ------------------------------------------------------------------------
    // --

    /**
     * Component reference
     */
    @Service
    Component weaveableC;

    /**
     * Logger of the component
     */
    @DreamMonolog()
    private Logger logger;

    /**
     * @see Push#push(Message)
     */
    public void push(Message message) throws PushException {
        ExportIdentifier exportidentifier = null;
        try {
            exportidentifier = this.destinationResolverItf.resolveDestination(message);
            this.lowerOutgoingPush.outgoingPush(message, exportidentifier);
        } catch (final InvalidExportIdentifierException e) {
            throw new PushException("Invalid destination for message", e);
        } catch (IOPushException e1) {
            // if it is a broken pipe, attempt a retry
            Throwable cause = e1.getCause();
            while ((cause) != null) {
                if ("Broken pipe".equals(cause.getMessage())) {
                    try {
                        try {
                            Thread.sleep(1000);
                        } catch (InterruptedException e) {
                            // do nothing
                        }
                        this.lowerOutgoingPush.outgoingPush(message, exportidentifier);
                        break;
                    } catch (InvalidExportIdentifierException e) {
                        throw new PushException("Invalid destination for message", e);
                    }
                }
                cause = cause.getCause();
            }
            if (cause == null) {
                throw e1;
            }
        }
    }

    /**
     * @see ChannelInExportId#getExportIdentifier()
     */
    public ExportIdentifier getExportIdentifier() {
        return this.lowerOutgoingPush.getLocalExportIdentifier();
    }

    // ------------------------------------------------------------------------
    // ---
    // Lifecycle methods
    // ------------------------------------------------------------------------
    // ---

    /**
     * Export the channel
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void startFcHandler() throws IllegalLifeCycleException {
        LoggerUtil.call(this.logger);

        try {
            this.lowerOutgoingPush = this.lowerProtocolItf.export(null, null);
        } catch (final ExportException e) {
            throw new IllegalLifeCycleException("Unable to export access point : " + e.getMessage());
        }
    }

    /**
     * Reset {@link #lowerOutgoingPush}
     */
    @LifeCycle(on = LifeCycleType.STOP)
    protected void stopFcHandler() throws IllegalLifeCycleException {
        this.lowerOutgoingPush = null;
    }
}
