/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.channel;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import org.objectweb.dream.control.activity.Util;
import org.objectweb.dream.control.activity.manager.TaskManager;
import org.objectweb.dream.control.activity.manager.ThreadPoolManager;
import org.objectweb.dream.control.activity.task.AbstractTask;
import org.objectweb.dream.control.activity.task.IllegalTaskException;
import org.objectweb.dream.control.activity.task.thread.ThreadPoolOverflowException;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamLifeCycle;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.dreamannotation.util.DreamLifeCycleType;
import org.objectweb.dream.protocol.IPExportIdentifier;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Implementation of {@link ConnectionFactoryItf}. Basically it contains a
 * multi-threaded listen task to create connections. There is one thread per
 * serverSocket.
 */
@DreamComponent(controllerDesc = "activeDreamPrimitive")
public class ConnectionFactory implements ConnectionFactoryItf {

    @SuppressWarnings("unused")
    private static final boolean DEFENSIVE_CHECK = true;

    /**
     * Registered Session Factories via the
     * {@link #addServerSocket(ServerSocket, IPExportIdentifier)} at export time
     */
    protected Map<IPExportIdentifier, SessionFactory> sessionFactories = new HashMap<IPExportIdentifier, SessionFactory>();

    /**
     * Available sessions.
     */
    protected BlockingQueue<SessionFactory> availableSessionFactories = new LinkedBlockingQueue<SessionFactory>();

    /**
     * Manager used to add/remove thread associated to the listen task.
     */
    protected ThreadPoolManager threadPoolManager;

    /**
     * The {@link ListenTask} of the component
     */
    protected ListenTask listenTask = new ListenTask();

    // ------------------------------------------------------------------------
    // ---
    // Client interfaces
    // ------------------------------------------------------------------------
    // ---

    @Requires(name = ConnectionManager.ITF_NAME)
    protected ConnectionManager connectionManagerItf;

    @Requires(name = "task-manager")
    protected TaskManager taskManagerItf;

    // ------------------------------------------------------------------------
    // --
    // Services interfaces
    // ------------------------------------------------------------------------
    // --

    /**
     * Component reference
     */
    @Service
    Component weaveableC;

    /**
     * Logger of the component
     */
    @DreamMonolog()
    protected Logger logger;

    // ------------------------------------------------------------------------
    // ---
    // Implementation of the ConnectionFactoryItf interface
    // ------------------------------------------------------------------------
    // ---

    /**
     * @see ConnectionFactoryItf#addServerSocket(ServerSocket,
     *      IPExportIdentifier)
     */
    public void addServerSocket(ServerSocket serverSocket, IPExportIdentifier identifier) {
        final SessionFactory sessionFactory = new SessionFactory(serverSocket, identifier);
        this.sessionFactories.put(identifier, sessionFactory);
        this.availableSessionFactories.add(sessionFactory);

        try {
            this.logger.log(BasicLevel.DEBUG, "add a thread in thread pool, " + sessionFactory);
            this.threadPoolManager.addThread(this.listenTask);
        } catch (final ThreadPoolOverflowException e) {
            this.logger.log(BasicLevel.WARN, "Unable to add listen thread", e);
        } catch (final IllegalTaskException e) {
            Error.bug(this.logger, e);
        }
    }

    /**
     * @see ConnectionFactoryItf#removeServerSocket(IPExportIdentifier)
     */
    public synchronized void removeServerSocket(IPExportIdentifier identifier)
            throws InvalidExportIdentifierException {
        final SessionFactory sessionFactory = this.sessionFactories.remove(identifier);
        if (sessionFactory == null) {
            throw new InvalidExportIdentifierException("Unknown export identifier.");
        }
        try {
            sessionFactory.closeSession();
        } catch (final IOException e) {
            this.logger.log(BasicLevel.WARN, "exception caught while closing server socket", e);
        }

        // remove session factory from available list
        this.availableSessionFactories.remove(sessionFactory);
    }

    // ------------------------------------------------------------------------
    // ---
    // Inner Class
    // ------------------------------------------------------------------------
    // ---

    /**
     * A Session Factory wraps a server socket. This session factory is
     * identified by its {@link IPExportIdentifier}
     * 
     * @author pgarcia
     */
    protected class SessionFactory {
        /**
         * <code>true</code> if the session factory is closed, i.e. the server
         * socket is closed.
         */
        private boolean closed = false;

        /**
         * The server socket on which accept() call will be made
         */
        private final ServerSocket serverSocket;

        /**
         * The identifier associated with the session factory
         */
        private final IPExportIdentifier exportIdentifier;

        protected SessionFactory(ServerSocket socket, IPExportIdentifier identifier) {
            this.serverSocket = socket;
            this.exportIdentifier = identifier;
        }

        protected Socket listen() throws IOException {
            return this.serverSocket.accept();
        }

        synchronized void closeSession() throws IOException {
            this.closed = true;
            this.serverSocket.close();
        }

        @Override
        public String toString() {
            return this.getClass().getSimpleName() + "@" + Integer.toHexString(this.hashCode())
                    + "(" + this.exportIdentifier + ")";
        }
    }

    /**
     * Task that execute the following operations :
     * <ol>
     * <li>Take(= reserve) an available session factory</li>
     * <li>Listens for a connection to be made (= creation of a new socket). The
     * method blocks until a connection is made.</li>
     * <li>Signal the connection manager (i.e. the protocol) that a new
     * connection has been made</li>
     * <li>Mark the session factory as available</li>
     * </ol>
     * 
     * @author pgarcia
     */
    protected class ListenTask extends AbstractTask {

        public ListenTask() {
            super("TCP/IP listen task");
        }

        // --------------------------------------------------------------------
        // Implementation of the Task interface
        // --------------------------------------------------------------------

        /**
         * @param hints
         *            not used here
         * @see org.objectweb.dream.control.activity.task.Task#execute(Object)
         */
        public Object execute(Object hints) throws InterruptedException {

            final SessionFactory sessionFactory = ConnectionFactory.this.availableSessionFactories
                    .poll(5, TimeUnit.SECONDS);
            if (sessionFactory == null) {
                ConnectionFactory.this.logger.log(BasicLevel.WARN,
                        "TCP/IP listen task will be stopped : no available session factory");
                return STOP_EXECUTING;
            }

            Socket socket;
            try {
                socket = sessionFactory.listen();
            } catch (final IOException e) {
                if (!sessionFactory.closed) {
                    try {
                        sessionFactory.serverSocket.close();
                    } catch (final IOException e1) {
                        // ignore
                    }
                    ConnectionFactory.this.logger.log(BasicLevel.ERROR,
                            "TCP/IP listen task will be stopped : " + e.getMessage());
                }

                return STOP_EXECUTING;
            }

            ConnectionFactory.this.connectionManagerItf.newConnection(socket,
                    sessionFactory.exportIdentifier);
            if (!sessionFactory.closed) {
                ConnectionFactory.this.availableSessionFactories.add(sessionFactory);
            }
            return EXECUTE_AGAIN;
        }

        /**
         * @see org.objectweb.dream.control.activity.task.Task#registered(Object)
         */
        @Override
        public void registered(Object controlItf) {
            ConnectionFactory.this.threadPoolManager = (ThreadPoolManager) controlItf;
        }

    }

    // ------------------------------------------------------------------------
    // Dream Lifecycle methods
    // ------------------------------------------------------------------------

    /**
     * Add the TCP/IP listen task
     */
    @DreamLifeCycle(on = DreamLifeCycleType.FIRST_START)
    protected void beforeFirstStart(Component componentItf) throws IllegalLifeCycleException {
        try {
            this.logger.log(BasicLevel.DEBUG,
                    "First start of the component : TCP/IP listen task added");

            final Map<String, Object> hints = new HashMap<String, Object>();
            hints.put("thread", "pool");
            Util.addTask(componentItf, this.listenTask, hints);
        } catch (final Exception e) {
            throw new IllegalLifeCycleException("Can't add TCP/IP listen task");
        }
    }

    /**
     * Close every opened sockets
     * 
     * @see org.objectweb.dream.control.lifecycle.PrepareStopLifeCycleController#
     *      prepareStopFc()
     */
    @DreamLifeCycle(on = DreamLifeCycleType.PREPARE_TO_STOP)
    public synchronized void prepareStopFc() throws IllegalLifeCycleException {
        this.logger.log(BasicLevel.DEBUG, "Going to stop, close every session factory");

        for (final SessionFactory sf : this.sessionFactories.values()) {
            try {
                this.logger.log(BasicLevel.DEBUG, "Close server sockets " + sf);
                sf.closeSession();
            } catch (final IOException e) {
                // ignore
            }
        }
    }
}
