/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.messagePassing;

import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.IPExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;

/**
 * Defines constants used as hints key that can be given to a UDP protocol.
 */
@Interface(name = UDPProtocolItf.ITF_NAME)
public interface UDPProtocolItf extends MessagePassingProtocol {
    /**
     * Name of the interface
     */
    String ITF_NAME = "protocol";

    /**
     * This integer hint, used at export time, can specify the port of the UDP
     * socket.<br>
     * If the hint is not specified, the socket will be created with an
     * undetermined port.<br>
     * This hint can also be used with the {@link
     * org.objectweb.dream.protocol.Protocol#createExportIdentifier} method to
     * specify the port of the export identifier to create.
     */
    String PORT = "port";

    /**
     * This integer hint, used at export time, can specify the range of ports on
     * which the UDP protocol will try to open socket. <br>
     * If this hint is not specified and a port has been specified (with {@link
     * #PORT} key), only the specified port will be tried before aborting. <br>
     * if neither this hint nor the {@link #PORT} hint are specified, the UDP
     * protocol will create a socket with an undetermined port.<br>
     */
    String PORT_RANGE = "range";

    /**
     * Key name for value used by the {@link
     * org.objectweb.dream.protocol.Protocol#createExportIdentifier} method to
     * specify the address of the export identifier to create. The value
     * associated with this key can be an {@link java.net.InetAddress} or a
     * {@link String}.
     */
    String ADDRESS = "address";

    /**
     * Utility method. Creates UDP socket using the given port and range.
     * 
     * @param incomingPushItf
     *      the interface on which message received on the created access point
     *      will be passed.
     * @param port
     *      the port of the created UDP socket. <br>
     *      If null, create the UDP socket with an undetermined port.
     * @param range
     *      the range of ports on which the UDP protocol will try to open
     *      socket. <br>
     *      If null, only one port will be tried before aborting. If negative,
     *      range is <i>infinite</i>.
     * @param fromChunkName
     *      If not <code>null</code> a {@link
     *      org.objectweb.dream.protocol.ExportIdentifierChunk} will be added to
     *      every received messages (see {@link #FROM_CHUNK_NAME}).
     * @return An export identifier.
     * @throws ExportException
     *      if the export process fails.
     * @see MessagePassingProtocol#export(IncomingPush, java.util.Map)
     */
    MessagePassingOutgoingPush export(IncomingPush incomingPushItf, int port, int range,
            String fromChunkName) throws ExportException;

    /**
     * Utility method.
     * 
     * @see MessagePassingProtocol#createExportIdentifier(java.util.Map,
     *  org.objectweb.dream.protocol.ExportIdentifier[])
     */
    IPExportIdentifier createExportIdentifier(String hostName, int port);
}
