/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s):
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.messagePassing.overChannel;

import java.io.IOException;

import org.objectweb.dream.IOPushException;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.ExportIdentifierChunk;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.OutgoingPush;
import org.objectweb.dream.protocol.channel.CloseChunk;
import org.objectweb.dream.util.LoggerUtil;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Message Passing Protocol Session
 * 
 * @author pgarcia
 */
public class Session implements IncomingPush {

    /**
     * Chunk name of {@link ExportIdentifierChunk} send in first message of
     * newly opened channel session.
     */
    public static final String ID_ANNOUNCEMENT_CHUNK_NAME = "id-announcement";

    /**
     * Inherited Logger of the protocol component
     */
    protected Logger logger;

    /**
     * Interface for the management of dream messages.
     */
    protected MessageManagerType messageManagerItf;

    /**
     * Identifier of the remote destination
     */
    protected ExportIdentifier remoteExportIdentifier;

    /**
     * Interface on which message are sent to the lower protocol
     */
    protected OutgoingPush outgoingPush;

    /**
     * The session which will replace this session
     */
    protected Session replacingSession;

    /**
     * The manager which manage this session
     */
    protected SessionManager sessionManager;

    /**
     * Chunk factory for export identifier chunk
     */
    protected ChunkFactoryReference<ExportIdentifierChunk> exportIdChunkFactory;

    protected ChunkFactoryReference<CloseChunk> closeChunkFactory;

    public Session() {
        super();
    }

    /**
     * Common internal constructor
     */
    private Session(MessageManagerType messageManagerItf, SessionManager sessionManager) {
        this.messageManagerItf = messageManagerItf;
        this.sessionManager = sessionManager;

        exportIdChunkFactory = messageManagerItf.getChunkFactory(ExportIdentifierChunk.class);
    }

    /**
     * Constructor use for an anonymous session (from an IN message)
     * 
     * @param outgoingPush
     * @param sessionManager
     */
    protected Session(OutgoingPush outgoingPush, SessionManager sessionManager) {
        this(sessionManager.messageManagerItf, sessionManager);

        this.logger = sessionManager.logger;
        this.logger.log(BasicLevel.DEBUG, "Constructor of an anonymous session");

        this.outgoingPush = outgoingPush;
    }

    /**
     * Constructor use for an already identified session (from an OUT message)
     * 
     * @param remoteExportIdentifier
     * @param sessionManager
     */
    protected Session(ExportIdentifier remoteExportIdentifier, SessionManager sessionManager) {
        this(sessionManager.messageManagerItf, sessionManager);

        this.logger = sessionManager.logger;
        this.logger.log(BasicLevel.DEBUG, "Constructor of an identified session with "
                + remoteExportIdentifier);

        this.remoteExportIdentifier = remoteExportIdentifier;
    }

    /**
     * Initialize the session for the OUT message.
     * 
     * @param outgoingPush
     * @throws IOPushException
     */
    protected void init(OutgoingPush outgoingPush) throws IOPushException {
        LoggerUtil.start(this.logger);

        this.outgoingPush = outgoingPush;
        if (sessionManager.localExportIdentifier != null) {
            final Message message = this.messageManagerItf.createMessage();
            final ExportIdentifierChunk chunk = messageManagerItf.createChunk(exportIdChunkFactory);
            chunk.setExportIdentifier(sessionManager.localExportIdentifier);
            this.messageManagerItf.addChunk(message, ID_ANNOUNCEMENT_CHUNK_NAME, chunk);
            this.outgoingPush.outgoingPush(message);
        }

        LoggerUtil.end(this.logger);
    }

    protected void close() throws IOException {
        this.outgoingPush.outgoingClose(this);
    }

    // -----------------------------------------------------------------------
    // Implementation of the IncomingPush interface
    // -----------------------------------------------------------------------

    /**
     * @see IncomingPush#incomingClosed(Object, Exception)
     */
    public void incomingClosed(Object outgoingPush, Exception exception) {

        if (this.logger.isLoggable(BasicLevel.WARN)) {
            this.logger.log(BasicLevel.WARN, "Lower " + this + " closed unexpectedly cause : ",
                    exception);
        }

        // ignore the closing message if the
        if (!sessionManager.closed) {
            sessionManager.removeSession(this);
        }
    }

    /**
     * @see IncomingPush#incomingPush(Message)
     */
    public void incomingPush(Message message) throws PushException {
        LoggerUtil.start(this.logger);

        if (sessionManager.closed) {
            throw new IOPushException("Access point has been closed.");
        }

        // opening message ?
        final ExportIdentifierChunk remoteExportIdChunk = this.messageManagerItf.getChunk(message,
                ID_ANNOUNCEMENT_CHUNK_NAME);
        if (remoteExportIdChunk != null) {
            this.treatIncomingIdentifingMessage(message, remoteExportIdChunk);
            return;
        }

        // closing message ?
        final CloseChunk closeChunk = this.messageManagerItf.getChunk(message,
                CloseChunk.DEFAULT_NAME);
        if (closeChunk != null) {
            this.treatIncomingClosingMessage(message, closeChunk);
            return;
        }

        // finally an upper message
        final ExportIdentifierChunk chunk = messageManagerItf.createChunk(exportIdChunkFactory);
        chunk.setExportIdentifier(remoteExportIdentifier);
        if (sessionManager.fromChunkName != null) {
            this.messageManagerItf.addChunk(message, sessionManager.fromChunkName, chunk);
        }
        sessionManager.upperIncomingPushItf.incomingPush(message);

        LoggerUtil.end(this.logger);
    }

    // ------------------------------------------------------------------------
    // Utility methods
    // ------------------------------------------------------------------------

    private final void treatIncomingIdentifingMessage(Message message,
            ExportIdentifierChunk remoteIdChunk) throws IOPushException {

        remoteExportIdentifier = remoteIdChunk.getExportIdentifier();
        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.logger.log(BasicLevel.DEBUG, "Treat anonymous session to identify from : "
                    + remoteExportIdentifier);
        }

        sessionManager.anonymousSessionIdentified(remoteExportIdentifier, this);
        this.messageManagerItf.deleteMessage(message);
    }

    private final void treatIncomingClosingMessage(Message message, CloseChunk closeChunk) {

        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.logger.log(BasicLevel.DEBUG, "Treat session to close from : "
                    + remoteExportIdentifier);
        }

        sessionManager.removeSession(this);

        try {
            outgoingPush.outgoingClose(this);
        } catch (final IOException e) {
            this.logger.log(BasicLevel.WARN, "Exception caught while closing lower session", e);
        }

        this.messageManagerItf.deleteMessage(message);
    }

    /**
     * @return true if the session is anonymous
     */
    protected final boolean isAnonymous() {
        return remoteExportIdentifier == null;
    }

    /**
     * Nice format for output
     */
    @Override
    public String toString() {
        final StringBuilder res = new StringBuilder();
        if (remoteExportIdentifier == null) {
            res.append("AnonymousSession@").append(Integer.toHexString(this.hashCode()))
                    .append("(");
        } else {
            res.append("Session@").append(Integer.toHexString(this.hashCode())).append("(").append(
                    remoteExportIdentifier).append(", ");
        }
        res.append("ReplacingSession : ").append(this.replacingSession).append(")");
        return res.toString();
    }
}
