/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Pierre Garcia
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.messagePassing.overChannel;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.objectweb.dream.IOPushException;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.protocol.BindException;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.ExportIdentifierChunk;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.protocol.OutgoingPush;
import org.objectweb.dream.protocol.channel.ChannelFactory;
import org.objectweb.dream.protocol.channel.ChannelProtocol;
import org.objectweb.dream.protocol.channel.CloseChunk;
import org.objectweb.dream.protocol.messagePassing.MessagePassingOutgoingPush;
import org.objectweb.dream.util.LoggerUtil;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * The Session Manager of the MessagePassingOverChannel implementation.
 * 
 * @author pgarcia
 * @author rnaudin - EBMWebsourcing
 * 
 */
public class SessionManager implements MessagePassingOutgoingPush, ChannelFactory {
    /**
     * for debug information. The code should optimize the code when false
     */
    // private static final boolean DEFENSIVE_CHECKS = false;
    
    /**
     * Inherited Logger of the protocol component
     */
    protected Logger logger;

    /**
     * Interface for the management of dream messages.
     */
    protected MessageManagerType messageManagerItf;

    /**
     * Interface related to the export/bind pattern.
     */
    protected ChannelProtocol lowerProtocolItf;

    /**
     * Chunk factory to create a chunk containing an {@link ExportIdentifier}
     */
    protected ChunkFactoryReference<ExportIdentifierChunk> exportIdChunkFactory;

    /**
     * Chunk factory to send close request to a session
     */
    protected ChunkFactoryReference<CloseChunk> closeChunkFactory;

    /**
     * Manager for anonymous session
     */
    protected AnonymousSessionManager anonymousSessionManager;

    /**
     * Manager for identified session, i.e. linked to an export identifier
     */
    protected IdentifiedSessionManager identifiedSessionManager;

    /**
     * Boolean variable used to know the state of the access point
     * <ul>
     * <li><code>true</code> if the communication channel (interface) has been
     * exported</li>
     * <li><code>false</code> otherwise</li>
     * </ul>
     */
    protected Boolean closed = false;

    protected String fromChunkName;

    protected ExportIdentifier localExportIdentifier;

    /**
     * Upper incoming push to transmit message upper in the protocol stack
     */
    protected IncomingPush upperIncomingPushItf;

    /**
     * Map of Export Identifiers to synchronize with
     */
    final private List<ExportIdentifier> lockExportIds;

    final private Map<ExportIdentifier, Integer> lockExportIdsMap;

    /**
     * Default constructor
     * 
     * @param messagePassingOverChannelImpl
     * @param upperIncomingPushItf
     * @param fromChunkName
     */
    protected SessionManager(MessagePassingOverChannelImpl messagePassingOverChannelImpl,
            IncomingPush upperIncomingPushItf, String fromChunkName) {
        // inherited references

        this.lockExportIds = new Vector<ExportIdentifier>();
        this.lockExportIdsMap = new HashMap<ExportIdentifier, Integer>();
        this.logger = messagePassingOverChannelImpl.logger;
        this.messageManagerItf = messagePassingOverChannelImpl.messageManagerItf;
        this.lowerProtocolItf = messagePassingOverChannelImpl.lowerProtocolItf;

        this.anonymousSessionManager = new AnonymousSessionManager(this.logger);
        this.identifiedSessionManager = new IdentifiedSessionManager(messagePassingOverChannelImpl,
                this);

        this.upperIncomingPushItf = upperIncomingPushItf;
        this.fromChunkName = fromChunkName;
        if (this.fromChunkName != null) {
            this.exportIdChunkFactory = this.messageManagerItf
                    .getChunkFactory(ExportIdentifierChunk.class);
        }
        this.closeChunkFactory = this.messageManagerItf.getChunkFactory(CloseChunk.class);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.dream.protocol.channel.ChannelFactory#instantiate(org.objectweb
     * .dream.protocol.OutgoingPush)
     */
    public IncomingPush instantiate(OutgoingPush toClientPush) throws BindException {
        LoggerUtil.call(this.logger);

        if (this.closed) {
            throw new BindException("Access point has been closed");
        }

        final Session session = new Session(toClientPush, this);
        this.anonymousSessionManager.add(session);

        return session;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.dream.protocol.messagePassing.MessagePassingOutgoingPush
     * #getLocalExportIdentifier()
     */
    public ExportIdentifier getLocalExportIdentifier() {
        return this.localExportIdentifier;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.dream.protocol.messagePassing.MessagePassingOutgoingPush
     * #outgoingPush(org.objectweb.dream.message.Message,
     * org.objectweb.dream.protocol.ExportIdentifier)
     */
    public void outgoingPush(Message message, ExportIdentifier to)
            throws InvalidExportIdentifierException, IOPushException {
        LoggerUtil.start(this.logger);

        // look for an identified session in the related manager
        if (this.closed) {
            throw new IOPushException("Access point has been closed.");
        }

        if (!this.checkLocalOutgoingPush(message, to)) {
            if (this.logger.isLoggable(BasicLevel.DEBUG)) {
                this.logger.log(BasicLevel.DEBUG, "Send a message with export identifier '" + to
                        + "' and local identified '" + this.localExportIdentifier + "'");
            }

            Session session;

            this.acquireLock(to);
            try {
                // reserve the session or create a new one if it does not exist
                session = this.identifiedSessionManager.lookup(to);
                if (session == null) {
                    session = this.createAndBindNewSession(to);
                } else {
                    this.identifiedSessionManager.reserveSession(session);
                }
            } finally {
                this.releaseLock(to);
            }

            // push outgoing message
            try {
                session.outgoingPush.outgoingPush(message);
            } catch (final IOPushException e) {
                this.releaseSession(session, e);
               throw e;
            }

            // finally release the session
            this.releaseSession(session, null);
        }

        LoggerUtil.end(this.logger);
    }

    /*
     * /(non-Javadoc)
     * 
     * @see
     * org.objectweb.dream.protocol.messagePassing.MessagePassingOutgoingPush
     * #outgoingClose(org.objectweb.dream.protocol.IncomingPush)
     */
    public void outgoingClose(IncomingPush incomingPush) throws IOException {
        LoggerUtil.call(this.logger);

        this.closed = true;
        
        // wait arbitrary that the ongoing out push are processed
        short count = 0;
        while (this.lockExportIds.size() > 0 && count < 10) {
            try {
                Thread.sleep(1000);
                count++;
            } catch (final InterruptedException e) {
                count = 10;
                // do nothing
            }
        }

        this.identifiedSessionManager.closeAll();
        this.anonymousSessionManager.closeAll();
        this.closeLocalAccessPoint();
    }

    /**
     * @param session
     * @throws IOPushException
     */
    protected void anonymousSessionIdentified(ExportIdentifier exportIdentifier, Session session)
            throws IOPushException {
        LoggerUtil.start(this.logger);

        this.acquireLock(exportIdentifier);

        this.anonymousSessionManager.remove(session);

        try {
            final Session sessionInCache = this.identifiedSessionManager
                    .lookup(session.remoteExportIdentifier);
            if (sessionInCache == null) {
                this.logger.log(BasicLevel.DEBUG, "New remote destination");
                session.remoteExportIdentifier = exportIdentifier;
                this.identifiedSessionManager.bind(exportIdentifier, session);
            } else {
                this.logger.log(BasicLevel.DEBUG, "Remote destination '"
                        + session.remoteExportIdentifier
                        + "' already exist, replace the current session");

                sessionInCache.replacingSession = session;

                if (this.identifiedSessionManager.isInUse(sessionInCache)) {
                    this.logger.log(BasicLevel.DEBUG,
                            "Session in cache is in use, it will be closed and replaced later");
                } else {
                    this.closeSessionSilently(sessionInCache);
                    this.identifiedSessionManager.unbind(sessionInCache);
                    this.identifiedSessionManager.bind(session.remoteExportIdentifier, session);
                }
            }
        } finally {
            this.releaseLock(exportIdentifier);
        }

        LoggerUtil.end(this.logger);
    }

    /**
     * Remove a {@link Session} identified by a {@link ExportIdentifier}.
     * 
     * @param remoteExportIdentifier
     *            the identifier of the session. if <code>null</code>, the
     *            session is considered as anonymous.
     * @param sessionToRemove
     */
    protected final void removeSession(Session sessionToRemove) {
        LoggerUtil.call(this.logger);

        if (!this.identifiedSessionManager.unbind(sessionToRemove)) {
            this.anonymousSessionManager.remove(sessionToRemove);
        }
    }

    /**
     * Close the session properly.<br>
     * Send a close request and then close the underling protocol.
     * 
     * @param session
     */
    protected void closeSessionSilently(Session session) {
        // send the close request if possible        
        final Message closeMessage = this.messageManagerItf.createMessage();
        final CloseChunk closeChunk = this.messageManagerItf.createChunk(this.closeChunkFactory);
        this.messageManagerItf.addChunk(closeMessage, CloseChunk.DEFAULT_NAME, closeChunk);
        try {
            session.outgoingPush.outgoingPush(closeMessage);
        } catch (final IOPushException e) {
            this.logger.log(BasicLevel.WARN, "Failed to send a closing message", e);
        }
        this.messageManagerItf.deleteMessage(closeMessage);

        // close the underlying session if possible
        try {
            session.close();
        } catch (final IOException e) {
            this.logger.log(BasicLevel.WARN, e);
        }
    }

    /**
     * Acquire a lock to work on a given destination.
     * 
     * @param exportIdentifier
     */
    protected void acquireLock(ExportIdentifier exportIdentifier) {
        LoggerUtil.call(this.logger);

        ExportIdentifier exportIdentifierLocker = null;

        synchronized (this.lockExportIds) {
            if (this.lockExportIds.contains(exportIdentifier)) {
                final Integer currentCounter = this.lockExportIdsMap.get(exportIdentifier);
                exportIdentifierLocker = this.lockExportIds.get(this.lockExportIds.indexOf(exportIdentifier));
                this.lockExportIdsMap.put(exportIdentifier, new Integer(currentCounter.intValue() + 1));
            } else {
                this.lockExportIds.add(exportIdentifier);
                this.lockExportIdsMap.put(exportIdentifier, new Integer(1));
            }
        }

        if (exportIdentifierLocker != null) {
            synchronized (exportIdentifierLocker) {
                try {
                    exportIdentifierLocker.wait();
                } catch (final InterruptedException e) {
                    this.logger.log(BasicLevel.LEVEL_WARN,
                            "The locking on the session bound to the destination '"
                                    + exportIdentifier + "' has been interrupted");
                }
            }
        }
    }

    /**
     * Release a lock on a given destination.
     * 
     * @param exportIdentifier
     */
    protected void releaseLock(ExportIdentifier exportIdentifier) {
        ExportIdentifier exportIdentifierLocker = null;

        LoggerUtil.call(this.logger);

        synchronized (this.lockExportIds) {
            Integer currentCounter = this.lockExportIdsMap.get(exportIdentifier);
            currentCounter = new Integer(currentCounter.intValue() - 1);
            this.lockExportIdsMap.put(exportIdentifier, currentCounter);
            if (currentCounter > 0) {
                exportIdentifierLocker = this.lockExportIds.get(this.lockExportIds.indexOf(exportIdentifier));
            } else {
                this.lockExportIds.remove(exportIdentifier);
            }
        }

        if (exportIdentifierLocker != null) {
            synchronized (exportIdentifierLocker) {
                exportIdentifierLocker.notify();
            }
        }
    }

    /**
     * Check if the outgoing push is for a local access point. If so, push it
     * locally.
     * 
     * @param message
     * @param to
     * @return {@code true} if it is a local push
     * @throws IOPushException
     */
    private boolean checkLocalOutgoingPush(Message message, ExportIdentifier to)
            throws IOPushException {
        boolean result = false;
        if (this.localExportIdentifier != null && this.localExportIdentifier.equals(to)) {
            result = true;
            this.logger.log(BasicLevel.DEBUG, "Send a message locally");
            // local sending : the message is directly pushed up in the stack
            if (this.fromChunkName != null) {
                final ExportIdentifierChunk chunk = this.messageManagerItf.createChunk(this.exportIdChunkFactory);
                chunk.setExportIdentifier(this.localExportIdentifier);
                this.messageManagerItf.addChunk(message, this.fromChunkName, chunk);
            }
            try {
                this.upperIncomingPushItf.incomingPush(message);
            } catch (final PushException e) {
                throw new IOPushException(e.getMessage());
            }
        }

        return result;

    }

    /**
     * Create/Use an entry in the cache and proceed the bind step of the
     * export/bind pattern
     * 
     * @param to
     * @return
     * @throws InvalidExportIdentifierException
     * @throws IOPushException
     */
    private Session createAndBindNewSession(ExportIdentifier to)
            throws InvalidExportIdentifierException, IOPushException {
        LoggerUtil.start(this.logger);

        Session newSession = null;

        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.logger.log(BasicLevel.DEBUG, "Create new session to " + to);
        }

        newSession = new Session(to, this);
        OutgoingPush lowerOutgoingPush = null;

        try {
            lowerOutgoingPush = this.bindSession(to, newSession);
        } catch (final BindException e) {
            throw new IOPushException(e);
        }

        newSession.init(lowerOutgoingPush);

        this.identifiedSessionManager.bind(to, newSession);
        this.identifiedSessionManager.fix(newSession);

        LoggerUtil.end(this.logger);

        return newSession;
    }

    /**
     * Release the session and try to replace it.
     * 
     * @param session
     *            the session to release
     * @param exception
     *            the potential exception which occurred during the sent
     * @return the replacing session if it exist, <code>null</code> otherwise
     * @throws IOPushException
     *             if an error occurs while releasing the session
     */
    private void releaseSession(Session session, IOPushException exception) {
        LoggerUtil.start(this.logger);

        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            if (exception == null) {
                this.logger.log(BasicLevel.DEBUG, "Release session '" + session + "'");
            } else {
                this.logger.log(BasicLevel.DEBUG, "Release session with an exception '" + session
                        + "' (error=" + exception + ")");
            }
        }

        this.acquireLock(session.remoteExportIdentifier);

        final Session replaceSession = session.replacingSession;
        try {
            this.identifiedSessionManager.unfix(session);

            if (exception != null || replaceSession != null) {
                this.closeSessionSilently(session);
                this.identifiedSessionManager.unbind(session);
            }

        } finally {
            this.releaseLock(session.remoteExportIdentifier);
        }

        LoggerUtil.returnAfter(this.logger);
    }

    /**
     * Bind the session. bind represents here the bind step of the export/bind
     * pattern
     * 
     * @param to
     * @param ce
     * @return
     * @throws InvalidExportIdentifierException
     * @throws BindException
     */
    private OutgoingPush bindSession(ExportIdentifier to, Session session)
            throws InvalidExportIdentifierException, BindException {
        LoggerUtil.call(this.logger);

        OutgoingPush lowerOutgoingPush;
        if (this.upperIncomingPushItf == null) {
            lowerOutgoingPush = this.lowerProtocolItf.bind(to, null, null);
        } else {
            lowerOutgoingPush = this.lowerProtocolItf.bind(to, session, null);
        }

        return lowerOutgoingPush;
    }

    /**
     * Close the local access point.
     * @throws ExportException
     */
    private void closeLocalAccessPoint() throws ExportException {
        this.logger.log(BasicLevel.DEBUG, "Close local access point " + this.localExportIdentifier);

        if (this.localExportIdentifier != null) {
            try {
                this.lowerProtocolItf.unexport(this.localExportIdentifier);
            } catch (final InvalidExportIdentifierException e) {
                throw new ExportException("Unable to unexport previously exported channel factory",
                        e);
            }
        }
    }
}
