/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): Romain Lenglet
 */

package org.objectweb.dream.pushwithreturn;

import java.util.Iterator;

import org.objectweb.dream.InterruptedPushException;
import org.objectweb.dream.Push;
import org.objectweb.dream.Push1;
import org.objectweb.dream.PushException;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * This component has one {@link Push}input, one {@link Push}output, and one
 * {@link Push1}input.
 * <p>
 * It receives calls to the <code>push</code> method. It generates a key for
 * the message, then it pushes the message on its output. When a message with
 * the same key arrives on its <code>Push1</code> input, it triggers an
 * handler for this message.
 * <p>
 * <i>Note: </i> if a message with an "All" key is received on the input, it is
 * used as a return message of all the call to the <code>pushWithReturn</code>
 * method.
 * <p>
 * <i>Note 2: </i> if a return message does not correspond to a stored key, it
 * is forwarded.
 */
@DreamComponent(controllerDesc = "dreamPrimitive")
@Provides(interfaces = {
    @Interface(name = "in-push", signature = Push.class),
    @Interface(name = "in-push1", signature = Push1.class)})
public class SynchronizerAsynchronous implements Push, Push1
{
  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  /** The client interface used to access the message manager. */
  @Requires(name = "message-manager")
  MessageManagerType     messageManagerItf;

  /** The KeyGeneratorItf client interface. */
  @Requires(name = KeyGeneratorItf.ITF_NAME)
  KeyGeneratorItf        keyGeneratorItf;

  /** The WaitingKeysItf client interface. */
  @Requires(name = WaitingKeysItf.ITF_NAME)
  WaitingKeysItf         waitingKeysItf;

  /** The HandleReturnMessageItf client interface. */
  @Requires(name = HandleReturnMessageItf.ITF_NAME)
  HandleReturnMessageItf handleReturnMessageItf;

  /** The Push output interface. */
  @Requires(name = "outgoing-out-push")
  Push                   outgoingOutPushItf;

  /** The Push incoming interface. */
  @Requires(name = "incoming-out-push")
  Push                   incomingOutPushItf;

  static final Object    LOCK = new Object();

  // ---------------------------------------------------------------------------
  // Services interfaces
  // ---------------------------------------------------------------------------

  /**
   * Component reference
   */
  @Service
  Component              weaveableC;

  /**
   * Logger used in concret implementation
   */
  @DreamMonolog()
  protected Logger       logger;

  // ---------------------------------------------------------------------------
  // Implementation of the PushWithReturn interface.
  // ---------------------------------------------------------------------------

  /**
   * @see Push#push(Message)
   */
  public void push(Message message) throws PushException
  {
    Key key = null;
    try
    {
      // Generate a key for the message
      key = keyGeneratorItf.generateKey(message);
    }
    catch (Exception e)
    {
      logger.log(BasicLevel.ERROR, e);
      throw new PushException(e);
    }

    // Add the key to the HashMap
    WaitingKey waitingKey;
    synchronized (LOCK)
    {
      try
      {
        waitingKey = waitingKeysItf.createWaitingKey(key);
        // TODO Cloning policy should be set using an attribute
        messageManagerItf.duplicateMessage(message, false);
        waitingKey.setMessage(message);
      }
      catch (InterruptedException e)
      {
        throw new InterruptedPushException(e);
      }
    }

    // Push the message on the output
    outgoingOutPushItf.push(message);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Push1 interface.
  // ---------------------------------------------------------------------------

  /**
   * @see Push1#push1(Message)
   */
  public void push1(Message message) throws PushException
  {
    Key key = null;
    try
    {
      // Generate or retrieve a key for the message
      key = keyGeneratorItf.generateKey(message);
    }
    catch (Exception e)
    {
      logger.log(BasicLevel.ERROR, e);
      throw new PushException(e);
    }

    if (key != null)
    {
      if (key.isAll())
      {
        synchronized (LOCK)
        {
          Iterator iter;
          try
          {
            iter = waitingKeysItf.removeWaitingKeys(key);
          }
          catch (InterruptedException e)
          {
            throw new InterruptedPushException(e);
          }
          while (iter.hasNext())
          {
            WaitingKey waitingKeyTemp = (WaitingKey) iter.next();
            waitingKeyTemp.setReturnMessage(message);
            try
            {
              handleReturnMessageItf.handleReturnMessage(waitingKeyTemp);
            }
            catch (Exception e)
            {
              throw new PushException(e);
            }
          }
        }
      }
      else
      {
        try
        {
          synchronized (LOCK)
          {
            WaitingKey waitingKey = waitingKeysItf.removeWaitingKey(key);
            if (waitingKey != null)
            {
              waitingKey.setReturnMessage(message);
              try
              {
                handleReturnMessageItf.handleReturnMessage(waitingKey);
              }
              catch (Exception e)
              {
                throw new PushException(e);
              }
            }
            else
            {
              // Already received a key for this message
              messageManagerItf.deleteMessage(message);
            }
          }
        }
        catch (InterruptedException e)
        {
          throw new InterruptedPushException(e);
        }
      }
    }
    else
    {
      // No key --> forward the message
      incomingOutPushItf.push(message);
    }
  }
}
