/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): Romain Lenglet
 */

package org.objectweb.dream.pushwithreturn;

import java.util.Iterator;

import org.objectweb.dream.InterruptedPushException;
import org.objectweb.dream.Push;
import org.objectweb.dream.Push1;
import org.objectweb.dream.PushException;
import org.objectweb.dream.PushWithReturn;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Contingency;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * This component handles calls to the
 * {@link PushWithReturn#pushWithReturn(Message)}method (i.e. outgoing
 * messages). It has one {@link PushWithReturn}input, one {@link Push}output,
 * and one {@link Push1}input.
 * <p>
 * It receives calls to the <code>pushWithReturn</code> method. It generates a
 * key for the message, then it pushes the message on its output, and waits for
 * a message with the same key to arrive on its <code>Push1</code> input. This
 * message is the returned message of the <code>pushWithReturn</code> method.
 * <p>
 * <i>Note 1: </i> if a message with an "All" key is received on the input, it
 * is used as a return message of all the <code>pushWithReturn</code> method
 * calls currently waiting for a return message.
 * <p>
 * <i>Note 2: </i> if a return message does not correspond to a stored key, it
 * is forwarded.
 */
@DreamComponent(controllerDesc = "dreamPrimitive")
@Provides(interfaces = {
    @Interface(name = "in-push-with-return", signature = PushWithReturn.class),
    @Interface(name = "in-push1", signature = Push1.class)})
public class SynchronizerSynchronous implements PushWithReturn, Push1
{
  // --------------------------------------------------------------------------
  // Constant
  // --------------------------------------------------------------------------

  static final Object LOCK = new Object();

  // --------------------------------------------------------------------------
  // Client interfaces
  // --------------------------------------------------------------------------

  /** The client interface used to access the message manager. */
  @Requires(name = "message-manager")
  MessageManagerType  messageManagerItf;

  /** The KeyGeneratorItf client interface. */
  @Requires(name = KeyGeneratorItf.ITF_NAME)
  KeyGeneratorItf     keyGeneratorItf;

  /** The WaitingKeysItf client interface. */
  @Requires(name = WaitingKeysItf.ITF_NAME)
  WaitingKeysItf      waitingKeysItf;

  /** The Push output interface. */
  @Requires(name = "outgoing-out-push")
  Push                outgoingOutPushItf;

  /** The Push incoming interface. */
  @Requires(name = "incoming-out-push", contingency=Contingency.OPTIONAL)
  Push                incomingOutPushItf;

  // --------------------------------------------------------------------------
  // Services interfaces
  // --------------------------------------------------------------------------

  /**
   * Component reference
   */
  @Service
  Component           weaveableC;

  /**
   * Logger used in concret implementation
   */
  @DreamMonolog()
  protected Logger    logger;

  // ---------------------------------------------------------------------------
  // Implementation of the PushWithReturn interface.
  // ---------------------------------------------------------------------------

  /**
   * @see PushWithReturn#pushWithReturn(Message)
   */
  public Message pushWithReturn(Message message) throws PushException
  {
    return doPush(message);
  }

  /**
   * @see Push#push(Message)
   */
  public void push(Message message) throws PushException
  {
    doPush(message);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Push1 interface.
  // ---------------------------------------------------------------------------

  /**
   * @see Push1#push1(Message)
   */
  public void push1(Message message) throws PushException
  {
    Key key = null;
    try
    {
      // Generate or retrieve a key for the message
      key = keyGeneratorItf.generateKey(message);
    }
    catch (Exception e)
    {
      logger.log(BasicLevel.ERROR, e);
      throw new PushException(e);
    }

    if (key != null)
    {
      if (key.isAll())
      {
        synchronized (LOCK)
        {
          Iterator iter;
          try
          {
            iter = waitingKeysItf.removeWaitingKeys(key);
          }
          catch (InterruptedException e)
          {
            throw new InterruptedPushException(e);
          }
          while (iter.hasNext())
          {
            WaitingKey waitingKeyTemp = (WaitingKey) iter.next();
            waitingKeyTemp.setReturnMessage(message);
            synchronized (waitingKeyTemp)
            {
              waitingKeyTemp.setReturnMessageReceived(true);
              waitingKeyTemp.notifyAll();
            }
          }
        }
      }
      else
      {
        try
        {
          synchronized (LOCK)
          {
            WaitingKey waitingKey = waitingKeysItf.removeWaitingKey(key);
            if (waitingKey != null)
            {
              waitingKey.setReturnMessage(message);
              synchronized (waitingKey)
              {
                waitingKey.setReturnMessageReceived(true);
                waitingKey.notifyAll();
              }
            }
            else
            {
              // Already received a key for this message
              messageManagerItf.deleteMessage(message);
            }
          }
        }
        catch (InterruptedException e)
        {
          throw new InterruptedPushException(e);
        }
      }
    }
    else
    {
      // No key --> forward the message
      if (incomingOutPushItf != null)
      {
        incomingOutPushItf.push(message);
      }
      else
      {
        logger
            .log(
                BasicLevel.WARN,
                "Message has no key and incoming-out-push interface is not bound: delete the message");
        messageManagerItf.deleteMessage(message);
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Utility methods.
  // ---------------------------------------------------------------------------

  private Message doPush(Message message) throws PushException
  {
    Key key = null;
    try
    {
      // Generate a key for the message
      key = keyGeneratorItf.generateKey(message);
    }
    catch (Exception e)
    {
      logger.log(BasicLevel.ERROR, e);
      throw new PushException(e);
    }

    // Add the key to the HashMap
    WaitingKey waitingKey;
    synchronized (LOCK)
    {
      try
      {
        waitingKey = waitingKeysItf.createWaitingKey(key);
      }
      catch (InterruptedException e)
      {
        throw new InterruptedPushException(e);
      }
    }

    // Push the message on the output
    outgoingOutPushItf.push(message);

    // Wait for the return message
    Message returnMessage;
    synchronized (waitingKey)
    {
      try
      {
        if (!waitingKey.getReturnMessageReceived())
        {
          waitingKey.wait();
        }
      }
      catch (InterruptedException e)
      {
        throw new InterruptedPushException(e);
      }
      returnMessage = waitingKey.getReturnMessage();
    }
    return returnMessage;
  }
}
