/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Romain Lenglet
 */

package org.objectweb.dream.queue.buffer;

import java.lang.ref.Reference;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.message.Message;
import org.objectweb.dream.pool.Recyclable;
import org.objectweb.dream.queue.MessageKeyManager;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.Attribute;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * This buffer sorts messages using a key provided by a
 * {@link MessageKeyManager}. For a given key, messages are returned in a FIFO
 * order, and in mutual exclusion.
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class BufferKeyMutexSorted extends AbstractBuffer
    implements
      UnlockKey
{

  // ------------------------------------------------------------------------
  // Constants
  // ------------------------------------------------------------------------

  private static final boolean DEFENSIVE_CHECKS = true;

  // ------------------------------------------------------------------------
  // Attributes
  // ------------------------------------------------------------------------
  /**
   * The maximal capacity of the buffer. Inherited from {@link AbstractBuffer}
   */
  @Attribute(argument = "maxCapacity")
  protected int                maxCapacity            = 0;  
  
  // ------------------------------------------------------------------------
  // Fields
  // ------------------------------------------------------------------------

  // TODO use object pool for MessageList and ListElem
  Map<Object, MessageList>     map              = new HashMap<Object, MessageList>();
  MessageList                  head;
  MessageList                  tail;
  Object                       reusableKey;
  ThreadLocal<Object>          usedKey          = new ThreadLocal<Object>();

  // ---------------------------------------------------------------------------
  // Client interface.
  // ---------------------------------------------------------------------------
  @Requires(name = MessageKeyManager.ITF_NAME)
  MessageKeyManager            messageKeyManagerItf;
  
  // --------------------------------------------------------------------------
  // Services interfaces
  // --------------------------------------------------------------------------  
  
  /**
   * Component reference
   */
  @Service
  Component weaveableC;
  
  /**
   * Logger of the component
   */ 
  @DreamMonolog()
  protected Logger                   logger;  

  // ---------------------------------------------------------------------------
  // Implementation of the AbstractBuffer methods.
  // ---------------------------------------------------------------------------

  /**
   * @see AbstractBuffer#doAdd(Message)
   */
  protected void doAdd(Message message)
  {
    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "Do add message : " + message);
    }
    if (reusableKey == null)
    {
      reusableKey = messageKeyManagerItf.createKey();
    }
    messageKeyManagerItf.fillKey(reusableKey, message);
    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "message key =" + reusableKey);
    }
    MessageList messageList = map.get(reusableKey);
    if (messageList == null)
    {
      // new key in the map.
      logger.log(BasicLevel.DEBUG, "Add new key in the map.");
      Object key = messageKeyManagerItf.duplicateKey(reusableKey);
      messageList = new MessageList();
      messageList.key = key;
      map.put(key, messageList);
      addAtHead(messageList);
    }
    else if (DEFENSIVE_CHECKS && (!messageList.key.equals(reusableKey)))
    {
      Error.bug(logger);
    }
    addMessageAtLast(messageList, message);
    logger.log(BasicLevel.DEBUG, "Message added");
  }

  /**
   * @see AbstractBuffer#doRemove()
   */
  protected Message doRemove()
  {
    logger.log(BasicLevel.DEBUG, "Removing message");
    if (DEFENSIVE_CHECKS && usedKey.get() != null)
    {
      Error.bug(logger);
    }
    MessageList messageList = removeTail();
    usedKey.set(messageList.key);
    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger
          .log(BasicLevel.DEBUG, "Return Message with key=" + messageList.key);
    }
    return removeMessageAtFirst(messageList);
  }

  /**
   * @see AbstractBuffer#doRemoveAll()
   */
  protected Message doRemoveAll()
  {
    Error.bug(logger, new UnsupportedOperationException(
        "This method is not implemented"));
    return null;
  }

  /**
   * @see AbstractBuffer#doGet()
   */
  protected Message doGet()
  {
    Error.bug(logger, new UnsupportedOperationException(
        "This method is not implemented"));
    return null;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the UnlockKey interface.
  // ---------------------------------------------------------------------------

  /**
   * @see UnlockKey#unlockKey()
   */
  public void unlockKey()
  {
    synchronized (lock)
    {
      Object key = usedKey.get();
      if (key == null)
      {
        throw new IllegalStateException(
            "The current thread has no associated key.");
      }
      if (logger.isLoggable(BasicLevel.DEBUG))
      {
        logger.log(BasicLevel.DEBUG, "Unlock key=" + key);
      }
      MessageList messageList = map.get(key);
      if (DEFENSIVE_CHECKS && (messageList == null))
      {
        Error.bug(logger);
      }
      if (messageList.first == null)
      {
        if (DEFENSIVE_CHECKS && (messageList.last != null))
        {
          Error.bug(logger);
        }
        map.remove(key);
        // release messageList;
        messageKeyManagerItf.deleteKey(key);
      }
      else
      {
        addAtHead(messageList);
      }
      usedKey.set(null);
    }
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  protected void addAtHead(MessageList list)
  {
    if (DEFENSIVE_CHECKS && (list == null))
    {
      Error.bug(logger);
    }
    if (DEFENSIVE_CHECKS && (list.nextList != null))
    {
      Error.bug(logger);
    }
    if (DEFENSIVE_CHECKS && (list.prevList != null))
    {
      Error.bug(logger);
    }
    if (head == null)
    {
      // list is empty
      if (DEFENSIVE_CHECKS && (tail != null))
      {
        Error.bug(logger);
      }
      head = list;
      tail = list;
    }
    else
    {
      if (DEFENSIVE_CHECKS && (head.prevList != null))
      {
        Error.bug(logger);
      }
      list.nextList = head;
      list.nextList.prevList = list;
      head = list;
    }
    incrementAvailableMessagesCount(1);
  }

  protected MessageList removeTail()
  {
    if (DEFENSIVE_CHECKS && (tail == null))
    {
      Error.bug(logger);
    }
    if (DEFENSIVE_CHECKS && (tail.nextList != null))
    {
      Error.bug(logger);
    }
    MessageList objectList = tail;
    if (tail.prevList == null)
    {
      // the list contains only one element.
      if (DEFENSIVE_CHECKS && (head != tail))
      {
        Error.bug(logger);
      }
      tail = null;
      head = null;
    }
    else
    {
      if (DEFENSIVE_CHECKS && (tail.prevList.nextList != tail))
      {
        Error.bug(logger);
      }
      tail = tail.prevList;
      tail.nextList = null;
      objectList.prevList = null;
    }
    incrementAvailableMessagesCount(-1);
    return objectList;
  }

  protected void addMessageAtLast(MessageList messageList, Message message)
  {
    ListElem listElem = new ListElem();
    listElem.message = message;
    if (messageList.last == null)
    {
      // list is empty
      if (DEFENSIVE_CHECKS && (messageList.first != null))
      {
        Error.bug(logger);
      }
      messageList.first = listElem;
    }
    else
    {
      if (DEFENSIVE_CHECKS && (messageList.first == null))
      {
        Error.bug(logger);
      }
      messageList.last.next = listElem;
    }
    messageList.last = listElem;
    incrementStoredMessagesCount(1);
  }

  protected Message removeMessageAtFirst(MessageList messageList)
  {
    if (DEFENSIVE_CHECKS && (messageList.first == null))
    {
      Error.bug(logger);
    }
    ListElem elem = messageList.first;
    messageList.first = elem.next;
    if (messageList.first == null)
    {
      // list contains only one element.
      if (DEFENSIVE_CHECKS && (messageList.last != elem))
      {
        Error.bug(logger);
      }
      messageList.last = null;
    }
    else if (DEFENSIVE_CHECKS && (messageList.last == null))
    {
      Error.bug(logger);
    }
    // release listElem
    incrementStoredMessagesCount(-1);
    return elem.message;
  }

  // ---------------------------------------------------------------------------
  // Utility classes
  // ---------------------------------------------------------------------------

  /** Object list */
  public static class MessageList implements Recyclable
  {
    Object                           key;
    ListElem                         first;
    ListElem                         last;
    MessageList                      nextList;
    MessageList                      prevList;
    Reference< ? extends Recyclable> recyclableReference;

    // -------------------------------------------------------------------------
    // Implementation of the Recyclable interface.
    // -------------------------------------------------------------------------

    /**
     * @see Recyclable#recycle()
     */
    public void recycle()
    {
      key = null;
      first = null;
      last = null;
      nextList = null;
      prevList = null;
    }

    /**
     * @see Recyclable#setReference(Reference)
     */
    public void setReference(
        Reference< ? extends Recyclable> recyclableReference)
    {
      this.recyclableReference = recyclableReference;
    }

    /**
     * @see Recyclable#getReference()
     */
    public Reference< ? extends Recyclable> getReference()
    {
      return recyclableReference;
    }
  }

  /** Object list elements. */
  public static class ListElem implements Recyclable
  {
    Message                          message;
    ListElem                         next;
    Reference< ? extends Recyclable> recyclableReference;

    // -------------------------------------------------------------------------
    // Implementation of the Recyclable interface.
    // -------------------------------------------------------------------------

    /**
     * @see Recyclable#recycle()
     */
    public void recycle()
    {
      message = null;
      next = null;
    }

    /**
     * @see Recyclable#setReference(Reference)
     */
    public void setReference(
        Reference< ? extends Recyclable> recyclableReference)
    {
      this.recyclableReference = recyclableReference;
    }

    /**
     * @see Recyclable#getReference()
     */
    public Reference< ? extends Recyclable> getReference()
    {
      return recyclableReference;
    }
  }
  
  // ---------------------------------------------------------------------------
  // Overriding of the AttributeController interface
  // generated by fraclet
  // ---------------------------------------------------------------------------

  /**
   * BufferMonitoring#setMaxCapacity(int)
   */
  public void setMaxCapacity(int maxCapacity)
  {
    synchronized (lock)
    {
      int previousCapacity = this.maxCapacity;
      this.maxCapacity = maxCapacity;
      if (maxCapacity > previousCapacity)
      {
        lock.notifyAll();
      }
    }
  }
}
