/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Romain Lenglet
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.queue.handler.incoming;

import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.queue.buffer.BufferAdd;
import org.objectweb.dream.queue.buffer.BufferOverflowException;
import org.objectweb.dream.queue.buffer.BufferRemove;

/**
 * An incoming handler that adds {@link #push(Message) pushed} messages into a
 * {@link BufferAdd buffer} bound to it. For each pushed message, this component
 * uses {@link BufferAdd#tryAdd(Message)} non blocking calls for trying to add
 * the pushed message into the buffer:
 * <ul>
 * <li>If the pushed message has been successfully added, everything is ok.
 * </li>
 * <li>If the pushed message cannot be added, an attempt is made to
 * {@link BufferRemove#tryRemove() remove} a message from the buffer.</li>
 * <ul>
 * <li>If this succeeds, the removed message is deleted, and another attempt is
 * made to {@link BufferAdd#tryAdd(Message) add} the pushed message into the
 * buffer.</li>
 * <ul>
 * <li>If this last attempt succeeds, everything is ok.</li>
 * <li>If this does not succeed, an {@link PushException exception} is thrown.
 * </li>
 * </ul>
 * <li>If this does not succeed, an {@link PushException exception} is thrown.
 * </li>
 * </ul>
 * </ul>
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class IncomingHandlerRemoving extends AbstractIncomingHandler
    implements
      Push
{

  // ------------------------------------------------------------------------
  // Client interfaces.
  // ------------------------------------------------------------------------
  
  

  // ------------------------------------------------------------------------
  // Implementation of the Push interface.
  // ------------------------------------------------------------------------

  /**
   * @see Push#push(Message)
   */
  public void push(Message message) throws PushException
  {
    if (!bufferAdd.tryAdd(message))
    {

      Message removedMessage = bufferRemove.tryRemove();
      if (removedMessage == null)
      {
        throw new BufferOverflowException(
            "unable to remove message from overflowed buffer");
      }

      messageManagerItf.deleteMessage(removedMessage);

      if (!bufferAdd.tryAdd(message))
      {
        throw new BufferOverflowException(
            "unable to add message into overflowed buffer");
      }

    }
  }

}
