/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Romain Lenglet
 * Contributor(s):
 */

package org.objectweb.dream.queue.keyed.buffer;

import org.objectweb.dream.message.Message;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;

/**
 * This interface defines methods that must be implemented by a buffer to allow
 * components to get or delete messages from it, according to two keys
 * associated with each message. Such an interface is used by <i>consumers </i>
 * of messages. <p/><i>Note 1: </i> all the messages that are stored in a
 * buffer are not always available for removal (e.g. when the buffer implements
 * an ordering policy). <p/><i>Note 2: </i> <b>buffer implementations MUST BE
 * thread-safe </b>.
 */
@Interface(name = DoubleKeyedBufferRemove.ITF_NAME)
public interface DoubleKeyedBufferRemove
{

  /**
   * The commonly used name to refer to an interface with this signature.
   */
  String ITF_NAME = "double-keyed-buffer-remove";

  /**
   * Gets the message from the buffer that is associated with the specified
   * keys. This method blocks while no message associated with the specified
   * keys is available.
   * 
   * @param key1 the first key that is associated with the message to get; never
   *          <code>null</code>.
   * @param key2 the second key that is associated with the message to get; can
   *          be <code>null</code>.
   * @return the message that is associated with the specified key; never
   *         <code>null</code>.
   * @throws InterruptedException if it is interrupted while waiting for a
   *           message to be added.
   * @see org.objectweb.dream.queue.buffer.BufferRemove#get()
   */
  Message get(Object key1, Object key2) throws InterruptedException;

  /**
   * Gets the message from the buffer that is associated with the specified
   * keys. This method does not block.
   * 
   * @param key1 the first key that is associated with the message to get; never
   *          <code>null</code>.
   * @param key2 the second key that is associated with the message to get; can
   *          be <code>null</code>.
   * @return the message that is associated with the specified keys if there is
   *         one, <code>null</code> otherwise.
   * @throws InterruptedException if it is interrupted while waiting for the
   *           mutex to be acquired.
   * @see org.objectweb.dream.queue.buffer.BufferRemove#tryGet()
   */
  Message tryGet(Object key1, Object key2) throws InterruptedException;;

  /**
   * Removes the message from the buffer that is associated with the specified
   * keys. This method blocks while no message associated with the specified
   * keys is available.
   * 
   * @param key1 the first key that is associated with the message to get; never
   *          <code>null</code>.
   * @param key2 the second key that is associated with the message to get; can
   *          be <code>null</code>.
   * @return the removed message that was associated with the specified keys;
   *         never <code>null</code>.
   * @throws InterruptedException if it is interrupted while waiting for a
   *           message to be added.
   * @see org.objectweb.dream.queue.buffer.BufferRemove#remove()
   */
  Message remove(Object key1, Object key2) throws InterruptedException;

  /**
   * Removes the message from the buffer that is associated with the specified
   * keys. This method does not block.
   * 
   * @param key1 the first key that is associated with the message to get; never
   *          <code>null</code>.
   * @param key2 the second key that is associated with the message to get; can
   *          be <code>null</code>.
   * @return the removed message that was associated with the specified key if
   *         there was one, <code>null</code> otherwise.
   * @see org.objectweb.dream.queue.buffer.BufferRemove#tryRemove()
   */
  Message tryRemove(Object key1, Object key2);

  /**
   * Removes all the available messages from the buffer that maps the specified
   * key. Messages are returned in an aggregated message. If no message has been
   * removed, this method returns <code>null</code>. This method does not
   * block.
   * 
   * @param key the key which messages that are removed must match.
   * @return an aggregated message containing all the messages available from
   *         the buffer, or <code>null</code> if no message was available.
   * @see org.objectweb.dream.queue.buffer.BufferRemove#removeAll()
   */
  Message removeAll(Object key);

  /**
   * Removes all the available messages from the buffer. Messages are returned
   * in an aggregated message. If no message has been removed, this method
   * returns <code>null</code>. This method does not block.
   * 
   * @return an aggregated message containing all the messages available from
   *         the buffer, or <code>null</code> if no message was available.
   * @see org.objectweb.dream.queue.buffer.BufferRemove#removeAll()
   */
  Message removeAll();

}