/**
 * Dream
 * Copyright (C) 2003 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Pierre Garcia
 * 
 */

/**
 * High performance implementation of Keyed Queue of Push/Pull Fast queues. 
 * Queues have the following properties :
 * <ul>
 * <li>FIFO</li>
 * <li>Thread safe</li>
 * <li>Has no maximum capacity</li>
 * <li>{@link #pull} is blocking if queue is empty.</li>
 * </ul>
 */

package org.objectweb.dream.queue.keyed.queue;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.objectweb.dream.PullException;
import org.objectweb.dream.PushException;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.queue.PushPullQueueFast;
import org.objectweb.dream.queue.keyed.KeyedPull;
import org.objectweb.dream.queue.keyed.KeyedPush;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * A keyed queue of queue of message The type of queues contained in the queue
 * is {@link PushPullQueueFast}
 * 
 * @author garciap
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
@Provides(interfaces = {
    @Interface(name = "out-keyed-pull", signature = KeyedPull.class),
    @Interface(name = "in-keyed-push", signature = KeyedPush.class)})
public class KeyedQueuePushPullQueuesFast
    implements
      KeyedPull,
      KeyedPush,
      KeyedQueueOfQueuesMonitoring
{

  @Service
  Component                      weaveable;

  @DreamMonolog
  Logger                         log;

  /**
   * Queue of queues
   */
  Map<Object, PushPullQueueFast> queue = new HashMap<Object, PushPullQueueFast>();

  // --------------------------------------------------------------------------
  // Keyed Pull Interface implementation
  // --------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.queue.keyed.KeyedPull#pull(java.lang.Object)
   */
  public Message pull(Object key) throws PullException
  {
    log.log(BasicLevel.DEBUG, "Pulling messages with the key " + key);
    PushPullQueueFast pushPullQueue = getQueue(key);
    // blocking call
    return pushPullQueue.pull();
  }

  // --------------------------------------------------------------------------
  // Keyed Push Interface implementation
  // --------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.queue.keyed.KeyedPush#push(org.objectweb.dream.message.Message,
   *      java.lang.Object)
   */
  public void push(Message message, Object key) throws PushException
  {
    log.log(BasicLevel.DEBUG, "Pushing messages with the key " + key);    
    PushPullQueueFast pushPullQueue = getQueue(key);
    pushPullQueue.push(message);
  }

  // --------------------------------------------------------------------------
  // Buffer Keyed Monitoring Interface
  // implementation
  // --------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.queue.keyed.buffer.BufferMonitoringKeyed#getKeys()
   */
  public synchronized Object[] getKeys()
  {
    return queue.keySet().toArray();
  }

  /**
   * @see org.objectweb.dream.queue.buffer.BufferMonitoring#getCurrentSize()
   */
  public synchronized int getCurrentSize()
  {
    Iterator<PushPullQueueFast> i = queue.values().iterator();
    int res = 0;
    while (i.hasNext())
    {
      res += i.next().getCurrentSize();
    }
    return res;
  }

  /**
   * @see org.objectweb.dream.queue.keyed.queue.KeyedQueueOfQueuesMonitoring#addQueue(java.lang.Object)
   */
  public synchronized boolean addQueue(Object key)
  {
    PushPullQueueFast result = queue.get(key);
    if (result == null)
    {
      result = new PushPullQueueFast();
      queue.put(key, result);
      return true;
    }
    return false;
  }

  /**
   * @see org.objectweb.dream.queue.keyed.queue.KeyedQueueOfQueuesMonitoring#removeQueue(java.lang.Object)
   */
  public synchronized boolean removeQueue(Object key)
  {
    PushPullQueueFast result = queue.get(key);
    if (result != null)
    {
      if (result.getCurrentSize() == 0)
      {
        queue.remove(key);
        return true;
      }
      else
      {
        return false;
      }
    }
    return true;
  }

  /**
   * @see org.objectweb.dream.queue.keyed.queue.KeyedQueueOfQueuesMonitoring#getQueueSize(java.lang.Object)
   */
  public int getQueueSize(Object key)
  {
    PushPullQueueFast targetQueue = queue.get(key);
    if (targetQueue != null)
    {
      return targetQueue.getCurrentSize();
    }
    else
    {
      return 0;
    }
  }

  // --------------------------------------------------------------------------
  // Utility method
  // --------------------------------------------------------------------------

  /**
   * Fetch the queue matching the giving key
   */
  private synchronized PushPullQueueFast getQueue(Object key)
  {
    PushPullQueueFast result = queue.get(key);
    if (result == null)
    {
      result = new PushPullQueueFast();
      queue.put(key, result);
    }
    return result;
  }
}
