/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): Romain Lenglet
 */

package org.objectweb.dream.serializator;

import java.io.IOException;
import java.util.Iterator;

import org.objectweb.dream.IOPushException;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.message.AbstractChunk;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.message.UnknownChunkException;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;

/**
 * This component de-serializes messages it receives on its input and send them
 * on its output. Incoming messages must contain a {@link ByteArrayChunk} chunk
 * with its {@link ByteArrayChunk#DEFAULT_NAME default name}. If the incoming
 * message contains additional chunks, they are transfered in the de-serialized
 * message.
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
@Provides(interfaces = @Interface(name = "in-push", signature = Push.class))
public class PushPushDeSerializatorImpl implements Push {

    // ---------------------------------------------------------------------------
    // Client interfaces
    // ---------------------------------------------------------------------------

    @Requires(name = DeSerializatorItf.ITF_NAME)
    protected DeSerializatorItf<byte[]> deserializatorItf;

    @Requires(name = "message-manager")
    protected MessageManagerType messageManagerItf;

    @Requires(name = "out-push")
    protected Push outPushItf;

    // ---------------------------------------------------------------------------
    // Implementation of the Push interface
    // ---------------------------------------------------------------------------

    /**
     * @see Push#push(Message)
     */
    @SuppressWarnings("unchecked")
    public void push(Message message) throws PushException {
        ByteArrayChunk byteArrayChunk;

        try {
            byteArrayChunk = messageManagerItf.removeChunk(message, ByteArrayChunk.DEFAULT_NAME);
        } catch (UnknownChunkException e) {
            throw new PushException("Cannot find " + ByteArrayChunk.DEFAULT_NAME
                    + " ByteArrayChunk in incoming message");
        }

        // Unserialize and forward the message
        Message unserializedMessage = null;
        try {
            unserializedMessage = deserializatorItf.deserialize(byteArrayChunk.getByteArray());
        } catch (IOException e) {
            throw new IOPushException("Unable to unserialize the message" + e);
        }
        Iterator<String> chunkNames = messageManagerItf.getChunkNameIterator(message);
        while (chunkNames.hasNext()) {
            String chunkName = chunkNames.next();
            AbstractChunk chunk = messageManagerItf.removeChunk(message, chunkName);
            messageManagerItf.addChunk(unserializedMessage, chunkName, chunk);
        }
        messageManagerItf.deleteMessage(message);
        outPushItf.push(unserializedMessage);
    }
}
