/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq, Vivien Quema
 * Contributor(s):
 */

package org.objectweb.dream.synchro;


import org.objectweb.dream.util.EmptyStringArray;
import org.objectweb.dream.dreamannotation.DreamComponent;

/**
 * Basic implementation of the {@link MutexItf} interface: this mutex is
 * non-reentrant. This implementation is unsafe, i.e. the identity of components
 * acquiring/releasing the MutexItf is unknown. As a consequence, the acquired
 * mutex can be released by every component.
 * <p>
 * <b>Note: inspired by Doug Lea's implementation. </b>
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class Mutex  implements MutexItf
{

  /** The lock status. */
  protected boolean inuse = false;

  // ---------------------------------------------------------------------------
  // Implementation of the MutexItf interface.
  // ---------------------------------------------------------------------------

  /**
   * @see MutexItf#lock()
   */
  public void lock() throws InterruptedException
  {
    if (Thread.interrupted()) throw new InterruptedException();
    synchronized (this)
    {
      try
      {
        while (inuse)
          wait();
        inuse = true;
      }
      catch (InterruptedException ex)
      {
        notify();
        throw ex;
      }
    }
  }

  /**
   * @see MutexItf#timedLock(long)
   */
  public boolean timedLock(long msecs) throws InterruptedException
  {
    if (Thread.interrupted()) throw new InterruptedException();
    synchronized (this)
    {
      if (!inuse)
      {
        inuse = true;
        return true;
      }
      else if (msecs <= 0)
        return false;
      else
      {
        long waitTime = msecs;
        long start = System.currentTimeMillis();
        try
        {
          for (;;)
          {
            wait(waitTime);
            if (!inuse)
            {
              inuse = true;
              return true;
            }
            waitTime = msecs - (System.currentTimeMillis() - start);
            if (waitTime <= 0) return false;
          }
        }
        catch (InterruptedException ex)
        {
          notify();
          throw ex;
        }
      }
    }
  }

  /**
   * @see MutexItf#unlock()
   */
  public synchronized void unlock()
  {
    inuse = false;
    notify();
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return EmptyStringArray.EMPTY_STRING_ARRAY;
  }

}
